/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details:
 *
 * Copyright (C) 2008 - 2009 Novell, Inc.
 * Copyright (C) 2009 - 2010 Red Hat, Inc.
 * Copyright (C) 2011 Novatel Wireless, Inc.
 */

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "mm-modem-novatel-lte.h"
#include "mm-modem-lte-network.h"
#include "mm-modem-lte-card.h"
#include "mm-modem-simple.h"
#include "mm-options.h"
#include "mm-errors.h"
#include "mm-callback-info.h"
#include "mm-modem-helpers.h"

static void modem_init (MMModem *modem_class);
static void modem_simple_init (MMModemSimple *class);
static void modem_lte_network_init (MMModemLteNetwork *lte_network_class);
static void modem_lte_card_init (MMModemLteCard *lte_card_class);
static void do_connect (MMModem *modem,
                        const char *cmd,
                        MMModemFn callback,
                        gpointer user_data);


G_DEFINE_TYPE_EXTENDED (MMModemNovatelLte, mm_modem_novatel_lte, MM_TYPE_GENERIC_LTE, 0,
                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM, modem_init)
                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_SIMPLE, modem_simple_init)
                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_LTE_NETWORK, modem_lte_network_init)
                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_LTE_CARD, modem_lte_card_init))


MMModem *
mm_modem_novatel_lte_new (const char *device,
                          const char *driver,
                          const char *plugin)
{
    g_return_val_if_fail (device != NULL, NULL);
    g_return_val_if_fail (driver != NULL, NULL);
    g_return_val_if_fail (plugin != NULL, NULL);

    return MM_MODEM (g_object_new (MM_TYPE_MODEM_NOVATEL_LTE,
                                   MM_MODEM_MASTER_DEVICE, device,
                                   MM_MODEM_DRIVER, driver,
                                   MM_MODEM_PLUGIN, plugin,
                                   NULL));
}


/*****************************************************************************/
/* MMModemSimple interface */

typedef enum {
    SIMPLE_STATE_ENABLE = 0,
    SIMPLE_STATE_SET_TECH_ALLOWED,
    SIMPLE_STATE_SET_CALLPARM,
    SIMPLE_STATE_SET_PDNS,
    SIMPLE_STATE_SET_SDNS,
    SIMPLE_STATE_SET_PNBNS,
    SIMPLE_STATE_SET_SNBNS,
    SIMPLE_STATE_SET_APN,
    SIMPLE_STATE_SET_IP,
    SIMPLE_STATE_SET_AUTH,
    SIMPLE_STATE_SET_USER,
    SIMPLE_STATE_SET_PWD,
    SIMPLE_STATE_CONNECT,
    SIMPLE_STATE_DONE
} SimpleState;

/* Looks a value up in the simple connect properties dictionary.  If the
 * requested key is not present in the dict, NULL is returned.  If the
 * requested key is present but is not a string, an error is returned.
 */
static gboolean
simple_get_property (MMCallbackInfo *info,
                     const char *name,
                     GType expected_type,
                     const char **out_str,
                     guint32 *out_num,
                     gboolean *out_bool,
                     GError **error)
{
    GHashTable *properties = (GHashTable *) mm_callback_info_get_data (info, "simple-connect-properties");
    GValue *value;
    gint foo;

    g_return_val_if_fail (properties != NULL, FALSE);
    g_return_val_if_fail (name != NULL, FALSE);
    if (out_str)
        g_return_val_if_fail (*out_str == NULL, FALSE);

    value = (GValue *) g_hash_table_lookup (properties, name);
    if (!value)
        return FALSE;

    if ((expected_type == G_TYPE_STRING) && G_VALUE_HOLDS_STRING (value)) {
        *out_str = g_value_get_string (value);
        return TRUE;
    } else if (expected_type == G_TYPE_UINT) {
        if (G_VALUE_HOLDS_UINT (value)) {
            *out_num = g_value_get_uint (value);
            return TRUE;
        } else if (G_VALUE_HOLDS_INT (value)) {
            /* handle ints for convenience, but only if they are >= 0 */
            foo = g_value_get_int (value);
            if (foo >= 0) {
                *out_num = (guint) foo;
                return TRUE;
            }
        }
    } else if (expected_type == G_TYPE_BOOLEAN && G_VALUE_HOLDS_BOOLEAN (value)) {
        *out_bool = g_value_get_boolean (value);
        return TRUE;
    }

    g_set_error (error, MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
                 "Invalid property type for '%s': %s (%s expected)",
                 name, G_VALUE_TYPE_NAME (value), g_type_name (expected_type));

    return FALSE;
}

static const char *
simple_get_string_property (MMCallbackInfo *info, const char *name, GError **error)
{
    const char *str = NULL;

    simple_get_property (info, name, G_TYPE_STRING, &str, NULL, NULL, error);
    return str;
}

static void
connect_report_done (MMAtSerialPort *port,
                     GString *response,
                     GError *error,
                     gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    GError *real_error;

    /* If the CEER command was successful, copy that error reason into the
     * callback's error.  If not, use the original error.
     */

    /* Have to do this little dance since mm_generic_lte_connect_complete()
     * copies the provided error into the callback info.
     */
    real_error = info->error;
    info->error = NULL;

    if (   !error
        && g_str_has_prefix (response->str, "+CEER: ")
        && (strlen (response->str) > 7)) {
        /* copy the connect failure reason into the error */
        g_free (real_error->message);
        real_error->message = g_strdup (response->str + 7); /* skip the "+CEER: " */
    }

    mm_generic_lte_connect_complete (MM_GENERIC_LTE (info->modem), real_error, info);
    if(real_error)
        g_error_free (real_error);
}

static void
connect_status_done (MMAtSerialPort *port,
                     GString *response,
                     GError *error,
                     gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    GError *real_error;
    real_error = info->error;
    info->error = NULL;

    if (   !error 
        && g_str_has_prefix (response->str, "$NWQMISTATUS: ")
        && (strlen (response->str) > 14 )) {
        if (g_strrstr(response->str, "CONNECTED") ) 
            mm_generic_lte_connect_complete (MM_GENERIC_LTE (info->modem), NULL, info);
        else {                                           
            /* copy the connect failure reason into the error */
            g_free (real_error->message);
            real_error->message = g_strdup (response->str + 14); /* skip the "$NWQMISTATUS: " */
            mm_generic_lte_connect_complete (MM_GENERIC_LTE (info->modem), real_error, info);
        }
    }
    if(real_error)
        g_error_free (real_error);
}

static void
connect_done (MMAtSerialPort *port,
              GString *response,
              GError *error,
              gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;

    if (error) {
        info->error = g_error_copy (error);
        /* Try to get more information why it failed */
        mm_at_serial_port_queue_command (port, "+CEER", 3, connect_report_done, info);
    } else {
        mm_at_serial_port_queue_command (port, "$NWQMISTATUS", 3, connect_status_done, info);
    }
}

static void
do_connect (MMModem *modem,
            const char *cmd,
            MMModemFn callback,
            gpointer user_data)
{
    MMCallbackInfo *info;
    MMAtSerialPort *port = mm_generic_lte_get_at_port ( MM_GENERIC_LTE(modem),
                                                        MM_PORT_TYPE_PRIMARY);    

    info = mm_callback_info_new (modem, callback, user_data);

    mm_modem_set_state (modem, MM_MODEM_STATE_CONNECTING, MM_MODEM_STATE_REASON_NONE);
    mm_at_serial_port_queue_command (port, cmd, 3, connect_done, info);
}


static char*
addToConnectCmd(char* cmd, const char* str)
{
    if (str)
        cmd = g_strdup_printf ("%s,%s", cmd, str);
    else 
        cmd = g_strdup_printf ("%s,", cmd);   
    return cmd;
}

static void
simple_state_machine (MMModem *modem, GError *error, gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    const char *str=NULL;
    static char *cmd = NULL;
    SimpleState state = GPOINTER_TO_UINT (mm_callback_info_get_data (info, "simple-connect-state"));
    SimpleState next_state = state;
    gboolean done = FALSE;
      
    info->error = mm_modem_check_removed (modem, error);
    if (info->error)
        goto out;

    if (mm_options_debug ()) {
        GTimeVal tv;
        char *data_device = NULL;

        g_object_get (G_OBJECT (modem), MM_MODEM_DATA_DEVICE, &data_device, NULL);
        g_get_current_time (&tv);
        g_debug ("<%ld.%ld> (%s): simple connect state %d",
                 tv.tv_sec, tv.tv_usec, data_device, state);
        g_free (data_device);
    }
   
    switch (state) {
          
    case SIMPLE_STATE_ENABLE:
        next_state = SIMPLE_STATE_SET_TECH_ALLOWED;
        mm_modem_enable (modem, simple_state_machine, info);
        break;
    case SIMPLE_STATE_SET_TECH_ALLOWED:
        next_state = SIMPLE_STATE_SET_CALLPARM;
        str = simple_get_string_property (info, "tech", &info->error);
        // Prepare QMI Connect command
        if(str)
            cmd = g_strdup_printf ("$NWQMICONNECT=%s", str);
        else
            cmd = g_strdup_printf ("$NWQMICONNECT=");
               
        /* otherwise fall through as no allowed mode was sent */
    case SIMPLE_STATE_SET_CALLPARM:
        next_state = SIMPLE_STATE_SET_PDNS;
        str = simple_get_string_property (info, "call_param", &info->error);
        cmd = addToConnectCmd (cmd, str);
               
    case SIMPLE_STATE_SET_PDNS:
        next_state = SIMPLE_STATE_SET_SDNS;
        str = simple_get_string_property (info, "pdns", &info->error);
        cmd = addToConnectCmd (cmd, str);
                
    case SIMPLE_STATE_SET_SDNS:
        next_state = SIMPLE_STATE_SET_PNBNS;
        str = simple_get_string_property (info, "sdns", &info->error);
        cmd = addToConnectCmd (cmd, str);
        
    case SIMPLE_STATE_SET_PNBNS:
        next_state = SIMPLE_STATE_SET_SNBNS;
        str = simple_get_string_property (info, "pnbns", &info->error);
        cmd = addToConnectCmd (cmd, str);
        
    case SIMPLE_STATE_SET_SNBNS:
        next_state = SIMPLE_STATE_SET_APN;
        str = simple_get_string_property (info, "snbns", &info->error);
        cmd = addToConnectCmd (cmd, str);
       
    case SIMPLE_STATE_SET_APN:
        next_state = SIMPLE_STATE_SET_IP;
        str = simple_get_string_property (info, "apn", &info->error);
        cmd = addToConnectCmd (cmd, str);
        if (str) {
            mm_modem_lte_network_set_apn (MM_MODEM_LTE_NETWORK (modem), str, simple_state_machine, info);
            break;
        }    
       
    case SIMPLE_STATE_SET_IP:
        next_state = SIMPLE_STATE_SET_AUTH;
        str = simple_get_string_property (info, "ip", &info->error);
        cmd = addToConnectCmd (cmd, str);
        
    case SIMPLE_STATE_SET_AUTH:
        next_state = SIMPLE_STATE_SET_USER;
        str = simple_get_string_property (info, "auth", &info->error);
        cmd = addToConnectCmd (cmd, str);
           
    case SIMPLE_STATE_SET_USER:
        next_state = SIMPLE_STATE_SET_PWD;
        str = simple_get_string_property (info, "user", &info->error);
        cmd = addToConnectCmd (cmd, str);
        
    case SIMPLE_STATE_SET_PWD:
        next_state = SIMPLE_STATE_CONNECT;
        str = simple_get_string_property (info, "password", &info->error);
        cmd = addToConnectCmd (cmd, str);
        
        /* Fall through if no APN or no 'apn' property error */
    case SIMPLE_STATE_CONNECT:
        next_state = SIMPLE_STATE_DONE;
        do_connect (modem, cmd, simple_state_machine, info);
        break;
    case SIMPLE_STATE_DONE:
        done = TRUE;
        break;
    }

 out:
    if (info->error || done) {
        mm_callback_info_schedule (info);
        g_free(cmd);
    }
    else
        mm_callback_info_set_data (info, "simple-connect-state", GUINT_TO_POINTER (next_state), NULL);
}

static void printConnectProperties(MMModemSimple *simple, GHashTable *properties)
{
    GHashTableIter iter;
    gpointer key, value;
    GTimeVal tv;
    char *data_device;

    g_object_get (G_OBJECT (simple), MM_MODEM_DATA_DEVICE, &data_device, NULL);
    g_get_current_time (&tv);

    g_hash_table_iter_init (&iter, properties);
    while (g_hash_table_iter_next (&iter, &key, &value)) {
        char *val_str;

        val_str = g_strdup_value_contents ((GValue *) value);
        g_debug ("<%ld.%ld> (%s): %s => %s",
                tv.tv_sec, tv.tv_usec,
                data_device, (const char *) key, val_str);
        g_free (val_str);
        }
    g_free (data_device);
}

static void
simple_connect (MMModemSimple *simple,
                GHashTable *properties,
                MMModemFn callback,
                gpointer user_data)
{
    MMCallbackInfo *info;
    
    /* If debugging, print all the simple connect properties */

    if (mm_options_debug ())
        printConnectProperties(simple, properties);
        
    info = mm_callback_info_new (MM_MODEM (simple), callback, user_data);
    mm_callback_info_set_data (info, "simple-connect-properties", 
                               g_hash_table_ref (properties),
                               (GDestroyNotify) g_hash_table_unref);

    simple_state_machine (MM_MODEM (simple), NULL, info);
}

/*****************************************************************************/
/*    Modem class override functions                                         */
/*****************************************************************************/

static void
dmat_callback2 (MMAtSerialPort *port,
                GString *response,
                GError *error,
                gpointer user_data)
{
    mm_serial_port_close (MM_SERIAL_PORT (port));
}

static void
dmat_callback (MMAtSerialPort *port,
               GString *response,
               GError *error,
               gpointer user_data)
{
    if (error) {
        /* Try it again */
        if (mm_serial_port_open (MM_SERIAL_PORT (port), NULL))
            mm_at_serial_port_queue_command (port, "$NWDMAT=1", 2, dmat_callback2, NULL);
    }
 
    mm_serial_port_close (MM_SERIAL_PORT (port));
}


static gboolean
grab_port (MMModem *modem,
           const char *subsys,
           const char *name,
           MMPortType suggested_type,
           gpointer user_data,
           GError **error)
{
    MMGenericLte *lte = MM_GENERIC_LTE (modem);
    MMPortType ptype = MM_PORT_TYPE_IGNORED;
    MMPort *port = NULL;
    
    if (suggested_type == MM_PORT_TYPE_UNKNOWN) {
        if (!mm_generic_lte_get_at_port (lte, MM_PORT_TYPE_PRIMARY)) 
                ptype = MM_PORT_TYPE_PRIMARY;
        else if (!mm_generic_lte_get_at_port (lte, MM_PORT_TYPE_SECONDARY))
            ptype = MM_PORT_TYPE_SECONDARY;
    } else
        ptype = suggested_type;

    port = mm_generic_lte_grab_port (lte, subsys, name, ptype, error);
    if (port && MM_IS_AT_SERIAL_PORT (port) && (ptype == MM_PORT_TYPE_PRIMARY)) {
        /* Novatel secondary port needs to be flipped from DM to AT mode
         * before it will answer our AT queries. So the primary port
         * needs this string first or auto detection of ctrl port fails.
         * Note: Early models/firmware were DM only
         */

        if (mm_serial_port_open (MM_SERIAL_PORT (port), NULL)) 
            mm_at_serial_port_queue_command (MM_AT_SERIAL_PORT (port), "$NWDMAT=1", 2, dmat_callback, NULL);
        
    }

    return !!port;
}

static void  disconnect_done (MMModem *modem,
                              GError *error,
                              gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    mm_callback_info_schedule (info);
}

static void
do_disconnect_done (MMAtSerialPort *port,
                    GString *response,
                    GError *error,
                    gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    MMModemState prev_state;
        
    info->error = mm_modem_check_removed (info->modem, error);
    if (info->error) {
     if(info->error->message)
        if (info->modem && info->modem) {
            /* Reset old state since the operation failed */
            prev_state = GPOINTER_TO_UINT (mm_callback_info_get_data (info, MM_GENERIC_LTE_PREV_STATE_TAG));
            mm_modem_set_state (MM_MODEM (info->modem),
                                prev_state,
                                MM_MODEM_STATE_REASON_NONE);
        }
    } else       
        mm_generic_lte_update_enabled_state (MM_GENERIC_LTE (info->modem), 
                                             FALSE, 
                                             MM_MODEM_STATE_REASON_NONE);
   
    mm_callback_info_schedule (info);
}


static void
disconnect_send_cmd (MMAtSerialPort *port,
                     gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    mm_at_serial_port_queue_command (MM_AT_SERIAL_PORT (port), "$NWQMIDISCONNECT", 2, do_disconnect_done, info);
}

static void
disconnect_flash_done (MMSerialPort *port,
                       GError *error,
                       gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    
    info->error = mm_modem_check_removed (info->modem, error);
    if (info->error) {
        /* Ignore "NO CARRIER" response when modem disconnects and any flash
         * failures we might encounter.  Other errors are hard errors.
         */
        if (   !g_error_matches (info->error, MM_MODEM_CONNECT_ERROR, MM_MODEM_CONNECT_ERROR_NO_CARRIER)
            && !g_error_matches (info->error, MM_SERIAL_ERROR, MM_SERIAL_ERROR_FLASH_FAILED)) {
            mm_callback_info_schedule (info);
            return;
        }
        g_clear_error (&info->error);
    }

    mm_port_set_connected (MM_PORT (port), FALSE);
    disconnect_send_cmd (MM_AT_SERIAL_PORT (port),
                         info);
    
}

static void
do_disconnect (MMModem *modem,
               MMModemFn callback,
               gpointer user_data)
{
    MMCallbackInfo *info;
    MMAtSerialPort *port = mm_generic_lte_get_at_port ( MM_GENERIC_LTE(modem),
                                                        MM_PORT_TYPE_PRIMARY);

    info = mm_callback_info_new (modem, callback, user_data);

    if (mm_serial_port_is_open (MM_SERIAL_PORT (port)))
            mm_serial_port_flash (MM_SERIAL_PORT (port), 1000, TRUE, disconnect_flash_done, info);
    else
       mm_callback_info_schedule (info);
}

static void
disconnect (MMModem *modem,
            MMModemFn callback,
            gpointer user_data)
{
    MMCallbackInfo *info;
    MMModemState state;
    info = mm_callback_info_new (modem, callback, user_data);

    /* Cache the previous state so we can reset it if the operation fails */
    state = mm_modem_get_state (modem);
    mm_callback_info_set_data (info,
                               MM_GENERIC_LTE_PREV_STATE_TAG,
                               GUINT_TO_POINTER (state),
                               NULL);
 
    mm_modem_set_state (modem, MM_MODEM_STATE_DISCONNECTING, MM_MODEM_STATE_REASON_NONE);
    do_disconnect (modem,  disconnect_done, info);
}

/*****************************************************************************/

static void
set_allowed_mode_done (MMAtSerialPort *port,
                       GString *response,
                       GError *error,
                       gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;

    if (error)
        info->error = g_error_copy (error);

   mm_callback_info_schedule (info);
}

static void
set_allowed_mode (MMGenericLte *lte,
                  MMModemLteAllowedMode mode,
                  MMModemFn callback,
                  gpointer user_data)
{
    MMCallbackInfo *info;
    MMAtSerialPort *port;
    char *command;
   
    info = mm_callback_info_new (MM_MODEM (lte), callback, user_data);

    port = mm_generic_lte_get_best_at_port (lte, &info->error);
    if (!port) {
        mm_callback_info_schedule (info);
        return;
    }

    command = g_strdup_printf ("$NWRAT=%d,2", mode);
    mm_at_serial_port_queue_command (port, command, 3, set_allowed_mode_done, info);
    g_free (command);
}

static gboolean
parse_nwrat_response (GString *response,
                      MMModemLteAllowedMode *out_mode,
                      GError **error)
{
    GRegex *r;
    GMatchInfo *match_info;
    char *str;
    gint mode = -1;
    gboolean success = FALSE;

    g_return_val_if_fail (response != NULL, FALSE);
    g_return_val_if_fail (out_mode != NULL, FALSE);

    r = g_regex_new ("\\$NWRAT:\\s*(\\d),(\\d),(\\d)", G_REGEX_UNGREEDY, 0, NULL);
    if (!r) {
        g_set_error_literal (error, MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
                             "Internal error parsing mode/tech response");
        return FALSE;
    }

    if (!g_regex_match_full (r, response->str, response->len, 0, 0, &match_info, NULL)) {
        g_set_error_literal (error, MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
                             "Failed to parse mode/tech response");
        goto out;
    }

    str = g_match_info_fetch (match_info, 1);
    mode = atoi (str);
    g_free (str);

    g_match_info_free (match_info);

    if (mode < 0 || mode > 2) {
        g_set_error_literal (error, MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
                             "Failed to parse mode/tech response");
        goto out;
    }
      
    if (out_mode) {
        if (mode == 0)
            *out_mode = MM_MODEM_LTE_ALLOWED_MODE_AUTOMATIC;
        else if (mode == 1)
            *out_mode = MM_MODEM_LTE_ALLOWED_MODE_GSM_PREFERRED;
        else if (mode == 2)
            *out_mode = MM_MODEM_LTE_ALLOWED_MODE_WCDMA_PREFERRED;
        else
            *out_mode = MM_MODEM_LTE_ALLOWED_MODE_AUTOMATIC;
    }
    success = TRUE;

out:
    g_regex_unref (r);
    return success;
}

static gboolean
parse_syscfg (const char *reply,
              guint32 *out_band)
{
    if (reply == NULL || strncmp (reply, "$NWBAND:", 8))
        return FALSE;

    if (sscanf (reply + 8, "%x", out_band))
        return TRUE;
    return FALSE;
}


static void
get_allowed_mode_done (MMAtSerialPort *port,
                       GString *response,
                       GError *error,
                       gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    MMModemLteAllowedMode mode = MM_MODEM_LTE_ALLOWED_MODE_AUTOMATIC;

    info->error = mm_modem_check_removed (info->modem, error);
    if (!info->error) {
        parse_nwrat_response (response, &mode, &info->error);
        mm_callback_info_set_result (info, GUINT_TO_POINTER (mode), NULL);
    }

    mm_callback_info_schedule (info);
}

static void
get_allowed_mode (MMGenericLte *lte,
                  MMModemUIntFn callback,
                  gpointer user_data)
{
    MMCallbackInfo *info;
    MMAtSerialPort *port;

    info = mm_callback_info_uint_new (MM_MODEM (lte), callback, user_data);

    port = mm_generic_lte_get_best_at_port (lte, &info->error);
    if (!port) {
        mm_callback_info_schedule (info);
        return;
    }

    mm_at_serial_port_queue_command (port, "$NWRAT?", 3, get_allowed_mode_done, info);
}

static void
set_band_done (MMAtSerialPort *port,
               GString *response,
               GError *error,
               gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    
    if (error)
        info->error = g_error_copy (error);
    
    mm_callback_info_schedule (info);
}

static void
set_band (MMModemLteNetwork *modem,
          guint32 band,
          MMModemFn callback,
          gpointer user_data)
{
    MMCallbackInfo *info;
    MMAtSerialPort *port;
    char *command;
   
    info = mm_callback_info_new (MM_MODEM (modem), callback, user_data);

    port = mm_generic_lte_get_best_at_port (MM_GENERIC_LTE (modem), &info->error);
    if (!port) {
        mm_callback_info_schedule (info);
        return;
    }

    mm_callback_info_set_data (info, "band", GUINT_TO_POINTER (band), NULL);
    command = g_strdup_printf ("AT$NWBAND=%X", band);
    mm_at_serial_port_queue_command (port, command, 3, set_band_done, info);
    g_free (command);
    
}

static void
get_band_done (MMAtSerialPort *port,
               GString *response,
               GError *error,
               gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
    guint32 band;
   
    if (error)
        info->error = g_error_copy (error);
    else if (parse_syscfg (response->str, &band)) 
        mm_callback_info_set_result (info, GUINT_TO_POINTER (band), NULL);
    
    mm_callback_info_schedule (info);
}

static void
get_band (MMModemLteNetwork *modem,
          MMModemUIntFn callback,
          gpointer user_data)
{
    MMAtSerialPort *port;
    MMCallbackInfo *info;

    info = mm_callback_info_uint_new (MM_MODEM (modem), callback, user_data);

    /* Prefer cached band from unsolicited messages if we have it */
    
    /* ask the modem */
    port = mm_generic_lte_get_best_at_port (MM_GENERIC_LTE (modem), &info->error);
    if (!port) {
        mm_callback_info_schedule (info);
        return;
    }

    mm_at_serial_port_queue_command (port, "AT$NWBAND?", 3, get_band_done, info);
}



static void
get_act_request_done (MMAtSerialPort *port,
                      GString *response,
                      GError *error,
                      gpointer user_data)
{
    MMCallbackInfo *info = user_data;
    MMModemLteAccessTech act = MM_MODEM_LTE_ACCESS_TECH_UNKNOWN;
    const char *p;

    if (error)
        info->error = g_error_copy (error);
    else {
        p = mm_strip_tag (response->str, "$CNTI:");
        p = strchr (p, ',');
        if (p)
            act = mm_lte_string_to_access_tech (p + 1);
    }

    mm_callback_info_set_result (info, GUINT_TO_POINTER (act), NULL);
    mm_callback_info_schedule (info);
}

static void
get_access_technology (MMGenericLte *modem,
                       MMModemUIntFn callback,
                       gpointer user_data)
{
    MMAtSerialPort *port;
    MMCallbackInfo *info;

    info = mm_callback_info_uint_new (MM_MODEM (modem), callback, user_data);
    port = mm_generic_lte_get_best_at_port (modem, &info->error);
    if (!port) {
        mm_callback_info_schedule (info);
        return;
    }

    mm_at_serial_port_queue_command (port, "$CNTI=0", 3, get_act_request_done, info);
}

static void
get_string_done (MMAtSerialPort *port,
                 GString *response,
                 GError *error,
                 gpointer user_data)
{
    MMCallbackInfo *info = (MMCallbackInfo *) user_data;

    if (error)
        info->error = g_error_copy (error);
    else
        mm_callback_info_set_result (info, g_strdup (response->str), g_free);

    mm_callback_info_schedule (info);
}

static void
get_imei (MMModemLteCard *modem,
          MMModemStringFn callback,
          gpointer user_data)
{
    MMAtSerialPort *port;
    MMCallbackInfo *info;

    info = mm_callback_info_string_new (MM_MODEM (modem), callback, user_data);
    port = mm_generic_lte_get_best_at_port (MM_GENERIC_LTE (modem), &info->error);
    mm_at_serial_port_queue_command_cached (port, "+GSN", 3, get_string_done, info);
}

/*****************************************************************************/

static void
modem_init (MMModem *modem_class)
{
    modem_class->grab_port  = grab_port;
    modem_class->disconnect = disconnect;
}

static void
modem_simple_init (MMModemSimple *class)
{
    class->connect = simple_connect;
}

static void
modem_lte_card_init (MMModemLteCard *class)
{
    class->get_imei = get_imei;   
}

static void
modem_lte_network_init (MMModemLteNetwork *class)
{
    class->set_band = set_band;
    class->get_band = get_band;
}

static void
mm_modem_novatel_lte_init (MMModemNovatelLte *self)
{
}

static void
mm_modem_novatel_lte_class_init (MMModemNovatelLteClass *klass)
{
    MMGenericLteClass *lte_class      = MM_GENERIC_LTE_CLASS (klass);
   
    mm_modem_novatel_lte_parent_class = g_type_class_peek_parent (klass);

    lte_class->set_allowed_mode       = set_allowed_mode;
    lte_class->get_allowed_mode       = get_allowed_mode;
    lte_class->get_access_technology  = get_access_technology;
}

